"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getPersistedConfig = exports.setPersistedConfig = exports.get$select = exports.restore$select = exports.hide$select = exports.mergeQueryFromAbility = exports.checkMulti = exports.getConditionalSelect = exports.handleConditionalSelect = exports.throwUnlessCan = exports.getAbility = exports.getAdapter = exports.makeDefaultOptions = exports.makeOptions = void 0;
const get_1 = __importDefault(require("lodash/get"));
const isEmpty_1 = __importDefault(require("lodash/isEmpty"));
const set_1 = __importDefault(require("lodash/set"));
const unset_1 = __importDefault(require("lodash/unset"));
const errors_1 = require("@feathersjs/errors");
const feathers_utils_1 = require("feathers-utils");
const getFieldsForConditions_1 = __importDefault(require("../../utils/getFieldsForConditions"));
const common_1 = require("../common");
const getAvailableFields_1 = __importDefault(require("../../utils/getAvailableFields"));
const hasRestrictingConditions_1 = __importDefault(require("../../utils/hasRestrictingConditions"));
const extra_1 = require("@casl/ability/extra");
const convertRuleToQuery_1 = __importDefault(require("../../utils/convertRuleToQuery"));
const makeOptions = (app, options) => {
    options = options || {};
    return Object.assign(common_1.makeDefaultBaseOptions(), defaultOptions, getAppOptions(app), options);
};
exports.makeOptions = makeOptions;
const defaultOptions = {
    adapter: undefined,
    availableFields: (context) => {
        var _a, _b;
        const availableFields = (_b = (_a = context.service.options) === null || _a === void 0 ? void 0 : _a.casl) === null || _b === void 0 ? void 0 : _b.availableFields;
        return getAvailableFields_1.default(context, { availableFields });
    }
};
const makeDefaultOptions = (options) => {
    return Object.assign(common_1.makeDefaultBaseOptions(), defaultOptions, options);
};
exports.makeDefaultOptions = makeDefaultOptions;
const getAppOptions = (app) => {
    const caslOptions = app === null || app === void 0 ? void 0 : app.get("casl");
    return (caslOptions && caslOptions.authorizeHook)
        ? caslOptions.authorizeHook
        : {};
};
const getAdapter = (context, options) => {
    var _a;
    if (options.adapter) {
        return options.adapter;
    }
    const caslAppOptions = (_a = context === null || context === void 0 ? void 0 : context.app) === null || _a === void 0 ? void 0 : _a.get("casl");
    if (caslAppOptions === null || caslAppOptions === void 0 ? void 0 : caslAppOptions.defaultAdapter) {
        return caslAppOptions.defaultAdapter;
    }
    return "feathers-memory";
};
exports.getAdapter = getAdapter;
const getAbility = (context, options) => {
    var _a, _b;
    // if params.ability is set, return it over options.ability
    if ((_a = context === null || context === void 0 ? void 0 : context.params) === null || _a === void 0 ? void 0 : _a.ability) {
        if (typeof context.params.ability === "function") {
            const ability = context.params.ability(context);
            return Promise.resolve(ability);
        }
        else {
            return Promise.resolve(context.params.ability);
        }
    }
    if (!options.checkAbilityForInternal && !((_b = context.params) === null || _b === void 0 ? void 0 : _b.provider)) {
        return Promise.resolve(undefined);
    }
    if (options === null || options === void 0 ? void 0 : options.ability) {
        if (typeof options.ability === "function") {
            const ability = options.ability(context);
            return Promise.resolve(ability);
        }
        else {
            return Promise.resolve(options.ability);
        }
    }
    if (exports.getPersistedConfig(context, "ability")) {
        if (typeof options.ability === "function") {
            const ability = options.ability(context);
            return Promise.resolve(ability);
        }
        else {
            return Promise.resolve(options.ability);
        }
    }
    return Promise.resolve(undefined);
};
exports.getAbility = getAbility;
const move = (context, fromPath, toPath) => {
    const val = get_1.default(context, fromPath);
    if (val !== undefined) {
        unset_1.default(context, fromPath);
        set_1.default(context, toPath, val);
    }
    return val;
};
const throwUnlessCan = (ability, method, resource, modelName, options) => {
    if (ability.cannot(method, resource)) {
        if (options.actionOnForbidden)
            options.actionOnForbidden();
        throw new errors_1.Forbidden(`You are not allowed to ${method} ${modelName}`);
    }
};
exports.throwUnlessCan = throwUnlessCan;
const handleConditionalSelect = (context, ability, method, modelName) => {
    var _a, _b;
    if (!((_b = (_a = context.params) === null || _a === void 0 ? void 0 : _a.query) === null || _b === void 0 ? void 0 : _b.$select)) {
        return false;
    }
    const { $select } = context.params.query;
    const $newSelect = exports.getConditionalSelect($select, ability, method, modelName);
    if ($newSelect) {
        exports.hide$select(context);
        context.params.query.$select = $newSelect;
        return true;
    }
    else {
        return false;
    }
};
exports.handleConditionalSelect = handleConditionalSelect;
const getConditionalSelect = ($select, ability, method, modelName) => {
    if (!($select === null || $select === void 0 ? void 0 : $select.length)) {
        return undefined;
    }
    const fields = getFieldsForConditions_1.default(ability, method, modelName);
    if (!fields.length) {
        return undefined;
    }
    const fieldsToAdd = fields.filter(field => !$select.includes(field));
    if (!fieldsToAdd.length) {
        return undefined;
    }
    return [...$select, ...fieldsToAdd];
};
exports.getConditionalSelect = getConditionalSelect;
const checkMulti = (context, ability, modelName, options) => {
    const { method } = context;
    const currentIsMulti = feathers_utils_1.isMulti(context);
    if (!currentIsMulti) {
        return true;
    }
    if ((method === "find" && ability.can(method, modelName)) ||
        (ability.can(`${method}-multi`, modelName))) {
        return true;
    }
    if (options === null || options === void 0 ? void 0 : options.actionOnForbidden)
        options.actionOnForbidden();
    throw new errors_1.Forbidden(`You're not allowed to multi-${method} ${modelName}`);
};
exports.checkMulti = checkMulti;
const adaptersFor$not = [
    "feathers-memory",
    "feathers-nedb",
    "feathers-objection",
    "feathers-sequelize"
];
const adaptersFor$nor = ["feathers-mongoose"];
const mergeQueryFromAbility = (context, ability, method, modelName, options) => {
    var _a;
    if (hasRestrictingConditions_1.default(ability, method, modelName)) {
        // TODO: if query and context.params.query differ -> separate calls
        const adapter = exports.getAdapter(context, options);
        let query;
        if (adaptersFor$not.includes(adapter)) {
            query = extra_1.rulesToQuery(ability, method, modelName, (rule) => {
                const { conditions } = rule;
                return (rule.inverted) ? { $not: conditions } : conditions;
            });
        }
        else if (adaptersFor$nor.includes(adapter)) {
            query = extra_1.rulesToQuery(ability, method, modelName, (rule) => {
                const { conditions } = rule;
                return (rule.inverted) ? { $nor: [conditions] } : conditions;
            });
        }
        else {
            query = extra_1.rulesToQuery(ability, method, modelName, (rule) => {
                const { conditions } = rule;
                return (rule.inverted) ? convertRuleToQuery_1.default(rule) : conditions;
            });
            if (query.$and) {
                const { $and } = query;
                delete query.$and;
                $and.forEach(q => {
                    query = feathers_utils_1.mergeQuery(query, q, {
                        defaultHandle: "intersect",
                        operators: context.service.operators,
                        useLogicalConjunction: true
                    });
                });
            }
        }
        if (!isEmpty_1.default(query)) {
            if (!context.params.query) {
                context.params.query = query;
            }
            else {
                const operators = (_a = context.service.options) === null || _a === void 0 ? void 0 : _a.whitelist;
                context.params.query = feathers_utils_1.mergeQuery(context.params.query, query, {
                    defaultHandle: "intersect",
                    operators,
                    useLogicalConjunction: true
                });
            }
        }
    }
};
exports.mergeQueryFromAbility = mergeQueryFromAbility;
const hide$select = (context) => {
    return move(context, "params.query.$select", "params.casl.$select");
};
exports.hide$select = hide$select;
const restore$select = (context) => {
    move(context, "params.casl.$select", "params.query.$select");
    return get_1.default(context, "params.query.$select");
};
exports.restore$select = restore$select;
const get$select = (context) => {
    return exports.getPersistedConfig(context, "$select");
};
exports.get$select = get$select;
const setPersistedConfig = (context, key, val) => {
    return set_1.default(context, `params.casl.${key}`, val);
};
exports.setPersistedConfig = setPersistedConfig;
const getPersistedConfig = (context, key) => {
    return get_1.default(context, `params.casl.${key}`);
};
exports.getPersistedConfig = getPersistedConfig;
